/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_hdlc.c,v 1.9 2005/11/29 22:29:31 sangoma Exp $
 */

/*****************************************************************************
 * sdla_chdlc.c	WANPIPE(tm) Multiprotocol WAN Link Driver. Cisco HDLC module.
 *
 * Author: 	Alex Feldman  <al.feldman@sangoma.com>
 *
 * ============================================================================
 * Nov 28, 2001 Alex Feldman		Added T1/E1 support (TE1).
 * Mar 10, 2000 Alex Feldman RX_CORR  o If receive buffer corrupted in
 *					rx_intr(), disable interrupt!!!
 * Dec 02, 1999 Alex Feldman          o Updated to FreeBSD 3.2
 * Aug 07, 1998	David Fong	      o Initial version.
 ****************************************************************************
 */
/*
****************************************************************************
****		For Debug purpose (only OpenBSD)			****
****************************************************************************
*/

#ifdef __OpenBSD__
# undef WANDEBUG_CHDLC	/* Uncomment this line for Debug output */
# undef ENABLE_IPV6	/* Uncomment this line for Enabling IPv6 support */
#endif

#include <net/wanpipe_includes.h>

#define _GNUC_
#include <net/wanpipe.h>		/* WANPIPE common user API definision */
#include <net/if_wanpipe_common.h>
#include <net/sdla_chdlc.h>		/* CHDLC firmware API definitions */
/****** Defines & Macros ****************************************************/

#ifdef	_DEBUG_
#define	STATIC
#define WANDEBUG_CHDLC
#else
#define	STATIC		static
#endif

/* reasons for enabling the timer interrupt on the adapter */
#define TMR_INT_ENABLED_UDP     0x0001
#define TMR_INT_ENABLED_UPDATE  0x0002
#define TMR_INT_ENABLED_TE  	0x0004
#define TMR_INT_ENABLED_CONFIG 	0x0008
#define TMR_INT_ENABLED_TEST 	0x0010

#define	CHDLC_DFLT_DATA_LEN	1500		/* default MTU */
#define CHDLC_HDR_LEN		1

#define IFF_POINTTOPOINT 0x10


#define	CHDLC_DFLT_DATA_LEN	1500		/* default MTU */
#define CHDLC_HDR_LEN		1

#define IFF_POINTTOPOINT 0x10

#define WANPIPE   0x00
#define API       0x01
#define CHDLC_API 0x01

#define PORT(x)  (x == 0 ? "PRIMARY" : "SECONDARY")

#define intr_permis flags.interrupt_info_struct.interrupt_permission
#define intr_type flags.interrupt_info_struct.interrupt_type

/******Data Structures*****************************************************/

/* This structure is placed in the private data area of the device structure.
 * The card structure used to occupy the private area but now the following 
 * structure will incorporate the card structure along with CHDLC specific data
 */

/* Route Status options */
#define NO_ROUTE	0x00
#define ADD_ROUTE	0x01
#define ROUTE_ADDED	0x02
#define REMOVE_ROUTE	0x03

/* Array of Sangoma card structure */
extern sdla_t sdla_card_array[];

#ifdef __OpenBSD__
extern int ticks;
#endif
/****** Function Prototypes *************************************************/

/* WAN link driver entry points. These are called by the WAN router module. */
static int update (wan_device_t* wandev);

static void chdlc_start(sdla_t*);

/* CHDLC Firmware interface functions */
static int set_adapter_config (sdla_t* card);
static int chdlc_configure 	(sdla_t* card, void* data);
static int chdlc_comm_enable 	(sdla_t* card);
static int chdlc_comm_disable 	(sdla_t* card);
static int chdlc_read_version 	(sdla_t* card, char* str);
static int chdlc_set_intr_mode 	(sdla_t* card, unsigned mode);
static int chdlc_send (sdla_t* card, void* data, unsigned int len);
#if 0
static int chdlc_read_intr_mode (sdla_t* card);
#endif
static int chdlc_read_comm_err_stats (sdla_t* card, COMMS_ERROR_STATS_STRUCT* stats);
static int chdlc_read_configure(sdla_t *card, CHDLC_CONFIGURATION_STRUCT* cfg);
static int chdlc_read_op_stats (sdla_t* card, CHDLC_OPERATIONAL_STATS_STRUCT* op_stats);

/* Miscellaneous CHDLC Functions */
static int set_chdlc_config (sdla_t *card);
static void init_chdlc_tx_rx_buff( sdla_t* card);
static int chdlc_error (sdla_t *card, int err, wan_mbox_t *mbox);
static int process_chdlc_exception(sdla_t *card);
static int process_global_exception(sdla_t *card);
static int update_comms_stats(sdla_t* card); 
static int configure_ip (sdla_t* card);
static int unconfigure_ip (sdla_t* card);
static int chdlc_config(sdla_t* card);

/* Interrupt handlers */
static void wpc_isr (sdla_t* card);
static void rx_intr (sdla_t* card);
static void timer_intr(sdla_t *);

#ifdef CHDLC_OLD_INTR_TEST
static int chdlc_intr_test(sdla_t* card);
#endif

/* TE1 */
static WRITE_FRONT_END_REG_T write_front_end_reg;
static READ_FRONT_END_REG_T  read_front_end_reg;
static void hdlc_enable_timer(void* card_id);
static void hdlc_handle_front_end_state(void* card_id);

/* Miscellaneous utilities function */
#ifdef WANDEBUG_CHDLC
static void chdlc_dump_mailbox(wan_mbox_t *);
#endif

extern struct mbuf* sdla_get(struct ifnet*, char*, unsigned);
static int  Intr_test_counter;

static int chdlc_debugging(sdla_t* card);
static unsigned long chdlc_crc_frames(sdla_t* card);
static unsigned long chdlc_abort_frames(sdla_t * card);
static unsigned long chdlc_tx_underun_frames(sdla_t* card);

#ifdef NETGRAPH
static int chdlc_up(sdla_t* card);
static void chdlc_down(sdla_t* card);
#endif /* NETGRAPH */

/****** Public Functions ****************************************************/

/* 
 * ============================================================================
 * Cisco HDLC protocol initialization routine.
 *
 * This routine is called by the main WANPIPE module during setup.  At this
 * point adapter is completely initialized and firmware is running.
 *  o read firmware version (to make sure it's alive)
 *  o configure adapter
 *  o initialize protocol-specific fields of the adapter data space.
 *
 * Return:	0	o.k.
 *		< 0	failure.
 */
int wph_init (sdla_t* card, wandev_conf_t* conf)
{
	CHDLC_CONFIGURATION_STRUCT	cfg;
	unsigned int 			timeout;
	unsigned char 			port_num;
	unsigned long 			max_permitted_baud = 0;
	union{
		char str[80];
	} u;
	int 	err, s = splimp();

	if (conf->config_id == WANCONFIG_MLINK_PPP){
#ifndef NETGRAPH
		DEBUG_EVENT("%s: Multi-Link PPP protocol is NETGRAPH based\n",
					card->devname);
		DEBUG_EVENT("%s: implementation.\n", card->devname);
		DEBUG_EVENT("%s: Enable NETGRAPH option in Makefile.in file!\n",
				card->devname);
		return -EINVAL;
#endif
	}
	/* Find out which Port to use */
	if ((conf->comm_port == WANOPT_PRI) || (conf->comm_port == WANOPT_SEC)){
		if (card->next){
			if (conf->comm_port != card->next->u.c.comm_port){
				card->u.c.comm_port = conf->comm_port;
			}else{
				DEBUG_EVENT("%s: %s port is already in use!\n",
						card->devname, 
						PORT(conf->comm_port));
				splx(s);
				return -EINVAL;
			}
		}else{
			card->u.c.comm_port = conf->comm_port;
		}
	}else{ 
		DEBUG_EVENT("%s: Invalid Port selected!\n", card->devname);
		splx(s);
		return -EINVAL;
	}
	
	/* Initialize protocol-specific fields */
	if (card->u.c.comm_port == WANOPT_PRI){
		card->mbox_off = PRI_BASE_ADDR_MB_STRUCT;
	}else{
		card->mbox_off = SEC_BASE_ADDR_MB_STRUCT;
	}

	port_num = card->u.c.comm_port;
	if (!card->wandev.piggyback){
		unsigned char	return_code;
	        /* The board will place an 'I' in the return code to indicate 
        	 * that it is ready to accept commands. We expect this to be
	         * completed in less than 1 second. 
        	 */
		timeout = SYSTEM_TICKS;
				do {
			return_code = 0x00;
			card->hw_iface.peek(card->hw, 
				card->mbox_off+offsetof(wan_mbox_t, wan_return_code),
				&return_code, 
				sizeof(unsigned char));
			if ((SYSTEM_TICKS - timeout) > 1*HZ) break;
		}while(return_code != 'I');
		if (return_code != 'I') {
			DEBUG_EVENT("%s: Initialization not completed by adapter\n", 
							card->devname);
        		DEBUG_EVENT("Please contact Sangoma representative.\n");
			splx(s);
			return -EIO;
		}
		DEBUG_INIT("%s: Received I from the board!!! Great\n",
							card->devname);
	}

	err = card->hw_iface.check_mismatch(card->hw,conf->fe_cfg.media);
	if (err){
		return err;
	}

	/* TE1 Make special hardware initialization for T1/E1 board */
	if (IS_TE1_MEDIA(&conf->fe_cfg)){
		
		memcpy(&card->fe.fe_cfg, &conf->fe_cfg, sizeof(sdla_fe_cfg_t));
		card->fe.name		= card->devname;
		card->fe.card		= card;
		card->fe.write_fe_reg	= write_front_end_reg;
		card->fe.read_fe_reg	= read_front_end_reg;
		card->wandev.fe_enable_timer = hdlc_enable_timer;
		card->wandev.te_link_state = hdlc_handle_front_end_state;
		conf->interface = 
			IS_T1_FEMEDIA(&card->fe) ? WANOPT_V35 : WANOPT_RS232;

		if (card->u.c.comm_port == WANOPT_PRI){
			conf->clocking = WANOPT_EXTERNAL;
		}

	}else if (IS_56K_MEDIA(&conf->fe_cfg)){

		memcpy(&card->fe.fe_cfg, &conf->fe_cfg, sizeof(sdla_fe_cfg_t));
		card->fe.name		= card->devname;
		card->fe.card		= card;
		card->fe.write_fe_reg	= write_front_end_reg;
		card->fe.read_fe_reg	= read_front_end_reg;
		
		if (card->u.c.comm_port == WANOPT_PRI){
			conf->clocking = WANOPT_EXTERNAL;
		}

	}else{
		card->fe.fe_status = FE_CONNECTED;
	}


	/* Read firmware version.  Note that when adapter initializes, it
	** clears the mailbox, so it may appear that the first command was
	** executed successfully when in fact it was merely erased. To work
	** around this, we execute the first command twice.
	*/

	if (chdlc_read_version(card, u.str)){
		splx(s);
		return -EIO;
	}
	if (conf->config_id == WANCONFIG_CHDLC){
		DEBUG_EVENT("%s: running Cisco HDLC firmware v%s\n", 
					card->devname, u.str); 
	}else if (conf->config_id == WANCONFIG_MLINK_PPP){
		DEBUG_EVENT("%s: Running Raw HDLC firmware v%s for Multi-Link PPP protocol.\n", 
					card->devname,u.str); 
	}else{
		DEBUG_EVENT("%s: Running Raw HDLC firmware v%s for Multi-Port PPP protocol.\n", 
					card->devname,u.str); 
	}

	if (set_adapter_config(card)) {
		splx(s);
		return -EIO;
	}

	card->isr			= &wpc_isr;
	card->poll			= NULL;
	card->exec			= NULL;
	card->wan_debugging		= &chdlc_debugging;
	card->get_crc_frames		= &chdlc_crc_frames;
	card->get_abort_frames		= &chdlc_abort_frames;
	card->get_tx_underun_frames	= &chdlc_tx_underun_frames;
	card->wandev.update		= &update;
	card->wandev.udp_port   	= conf->udp_port;

	/* Set the flag pointers for both ports to NULL */
	card->flags_off = 0;

	/* reset the number of times the 'update()' proc has been called */
	card->u.c.update_call_count = 0;
	
	card->wandev.ttl = conf->ttl;
	card->wandev.interface = conf->interface;
	
	if (card->u.c.comm_port == WANOPT_SEC && conf->interface == WANOPT_V35 &&
	    card->type != SDLA_S514){
		DEBUG_EVENT("%s: V35 Interface not supported on S508 %s Port!\n",
				card->devname, PORT(card->u.c.comm_port));
		splx(s);
		return -EIO;
	} 

	card->wandev.clocking = conf->clocking;
	
	if (card->wandev.clocking){
		if (port_num == WANOPT_PRI){ /* For primary port */
			max_permitted_baud = 
				(card->type == SDLA_S514) ? PRI_MAX_BAUD_RATE_S514 :
						       PRI_MAX_BAUD_RATE_S508;	
		}else{ 
			if (port_num == WANOPT_SEC){ /* For secondary port */
		        	max_permitted_baud = 
			    		(card->type == SDLA_S514) ? SEC_MAX_BAUD_RATE_S514 :
				  		           SEC_MAX_BAUD_RATE_S508;	
			}
		}
		if (conf->bps > max_permitted_baud){
			conf->bps = max_permitted_baud;
			DEBUG_EVENT("%s: Baud too high!!!\n", card->devname);
			DEBUG_EVENT("%s: Baud rate set to %lu bps.\n",
						card->devname, max_permitted_baud);
		}
		card->wandev.bps = conf->bps;
	}else{
		card->wandev.bps = 0;
	}
		
	/* Setup the Port MTU */
	if (port_num == WANOPT_PRI){ /* For primary port 0 */
		card->wandev.mtu = (conf->mtu >= MIN_LGTH_CHDLC_DATA_CFG) ?
				min(conf->mtu, PRI_MAX_NO_DATA_BYTES_IN_FRAME) :
				CHDLC_DFLT_DATA_LEN;
	}else{
		if (port_num == WANOPT_SEC){ /* For secondary port 1 */
			card->wandev.mtu = (conf->mtu >= MIN_LGTH_CHDLC_DATA_CFG) ?	
					min(conf->mtu, SEC_MAX_NO_DATA_BYTES_IN_FRAME) :
					CHDLC_DFLT_DATA_LEN;
		}
	}
	/* Set up the interrupt status area. 
	** Read the CHDLC_configuration and obtain:
	** Ptr to sahred memory info struct.
	** Use this pointer to calculate the value of card->u.c.flags!
	*/
	err = chdlc_read_configure(card, &cfg);
	if (err != WAN_CMD_OK){
		splx(s);
		return err;
	}

       	card->flags_off = cfg.ptr_shared_mem_info_struct;
	card->intr_type_off = card->flags_off + 
				offsetof(SHARED_MEMORY_INFO_STRUCT, interrupt_info_struct) +
			      	offsetof(INTERRUPT_INFORMATION_STRUCT, interrupt_type);
	card->intr_perm_off = card->flags_off +  
				offsetof(SHARED_MEMORY_INFO_STRUCT, interrupt_info_struct) +
				offsetof(INTERRUPT_INFORMATION_STRUCT, interrupt_permission);
	card->fe_status_off = card->flags_off + 
				offsetof(SHARED_MEMORY_INFO_STRUCT, FT1_info_struct) +
				offsetof(FT1_INFORMATION_STRUCT, parallel_port_A_input);

	/* This is for the ports link state */
	card->wandev.state	= WAN_DISCONNECTED;
	   
	splx(s);
#ifdef CHDLC_OLD_INTR_TEST
	if (!card->wandev.piggyback){
		err = chdlc_intr_test(card);
	}
#endif

	/* ALEX: MPD */
#ifdef NETGRAPH
	card->wan_up = &chdlc_up;
	card->wan_down = &chdlc_down;
	card->wan_start = &chdlc_start;
#endif
	card->u.c.protocol_options = HDLC_STREAMING_MODE;
	if (chdlc_set_intr_mode(card, APP_INT_ON_TIMER)){ 
		return -EIO;
	}
	if (!card->wandev.piggyback){
		Intr_test_counter = 0;
		card->timer_int_enabled |= TMR_INT_ENABLED_TEST;
		/* Enable watchdog timer in case if hardware i
		 * interrupt will failed */ 
	}else{
		card->timer_int_enabled |= TMR_INT_ENABLED_CONFIG;
	}

	return err;
}

/******* WAN Device Driver Entry Points *************************************/

/* 
 * ============================================================================
 * Update device status & statistics (called when updating the PROC file system)
 */
static int update (wan_device_t* wandev)
{
	sdla_t* 	card = NULL;
	unsigned long 	timeout;
 
	/* sanity checks */
	if ((wandev == NULL) || (wandev->private == NULL)){
		return -EFAULT;
	}

	if(wandev->state == WAN_UNCONFIGURED){
		return -ENODEV;
	}

	card = wandev->private;

	/* more sanity checks */
	if (!card->flags_off){
		return -ENODEV;
	}

	if (card->update_comms_stats){
		return -EAGAIN;
	}

	/* TE1_56K	
	 *  If update will need to read T1/E1/56K statistics,
	 *  change the update_comms_stats variable to 3, 
	 *  otherwise 2.
	 */
	card->update_comms_stats = 2; 
	
	card->hw_iface.set_bit(card->hw, card->intr_perm_off, APP_INT_ON_TIMER);
	card->timer_int_enabled = TMR_INT_ENABLED_UPDATE;

	/* wait maximum of 1 second for the statistics to be updated */
	timeout = SYSTEM_TICKS;
	for(;;){
		if (card->update_comms_stats == 0){
			break;
		}
		if ((SYSTEM_TICKS - timeout) > (1*hz)){
			card->update_comms_stats = 0;
			card->timer_int_enabled &= 
						~TMR_INT_ENABLED_UPDATE;
			return -EAGAIN;
		}
	}   
	return 0;
}


/* 
 * ============================================================================
 *
 *
 *
 */
static int chdlc_config (sdla_t* card)
{
	int	err = 0;

	DEBUG_INIT("%s: Configuring interface...\n", 
					card->devname);
  	if (set_chdlc_config(card)){ 
		return -EIO;
	}

	/* Initialize Rx/Tx buffer control fields */
	init_chdlc_tx_rx_buff(card);
	
	if (IS_TE1_CARD(card)) {
		DEBUG_EVENT("%s: Configuring onboard %s CSU/DSU\n",
			card->devname, 
			(IS_T1_CARD(card))?"T1":"E1");
		if (sdla_te_config(&card->fe, &card->wandev.fe_iface)){
			DEBUG_EVENT("%s: Failed %s configuratoin!\n",
						card->devname,
						(IS_T1_CARD(card))?"T1":"E1");
			return 0;
		}
	}

	if (IS_56K_CARD(card)) {
		DEBUG_EVENT("%s: Configuring 56K onboard CSU/DSU\n",
					card->devname);

		if(sdla_56k_config(&card->fe, &card->wandev.fe_iface)) {
			DEBUG_EVENT("%s: Failed 56K configuration!\n",
						card->devname);
			return 0;
		}
	}

	/* Disable All interrupts */
	chdlc_set_intr_mode(card, 0x00);

	wanpipe_set_state(card, WAN_CONNECTING);
	wan_getcurrenttime(&card->u.c.router_start_time, NULL);

	/* Manually poll the 56K CSU/DSU to get the status */
	if (IS_56K_CARD(card)){
		/* 56K Update CSU/DSU alarms */
		card->wandev.fe_iface.read_alarm(&card->fe, 1);
	}
	return err;
}

static void chdlc_start(sdla_t* card)
{
	int 		      	   	isempty = 0, s = 0;
	netskb_t*			opkt = NULL;	/* the packet */
	register netskb_t*		m = NULL;	/* current mbuf */
	unsigned short	           	data_length;	/* data length */
#ifdef ALTQ
	struct ifaltq*			tq = NULL;
#else
	struct ifqueue*			tq = NULL;
#endif

	s = splimp();

	/* We're trying to drain the queue as much as possible here because 
	** if we manage to get more than one packet into transmit queue     
	** ( line-down/line-up scenario, or heavy load ) and take only one  
	** at a time out we get an effective delay ...		      
	** BF-0001 
	*/ 
	while(1){
		isempty = 1;
#ifdef NETGRAPH
		tq = &card->xmitq_hipri;
		isempty = WAN_IFQ_IS_EMPTY(tq);
		if (isempty){
			tq = &card->xmitq;
			isempty = WAN_IFQ_IS_EMPTY(tq);
		}
		if (isempty){
			/* No more packets in send queue */
			break;
		}
		WAN_IFQ_POLL(tq, opkt);
#endif
		if (opkt == NULL){
			/* Should never happened, packet pointer is NULL */
			break;
		}
  
		/* 
		** ok, now when we've got an available buffer we can transmit something
		*/
		/* Copy the packet inside */
		data_length = 0;
		for(m = opkt; m != NULL; m = m->m_next){ 
			if (data_length + m->m_len > MAX_PACKET_SIZE){
				DEBUG_TX("%s: Dropt the packet (too big - %d bytes)!\n",
					card->devname, data_length);
				IF_DROP(tq);
		
				/* dequeue packet from IP queue */
				WAN_IFQ_DEQUEUE(tq, opkt);
				if (opkt){
					wan_skb_free(opkt);
				}
				continue;
			}
			bcopy(mtod(m, caddr_t), &card->tx_data[data_length], m->m_len);
			data_length += m->m_len;
		}
		card->tx_data[data_length] = '\0';	
		if (chdlc_send(card, card->tx_data, data_length)){

			/* enable transmit interrupts */
			card->hw_iface.set_bit(card->hw, card->intr_perm_off, APP_INT_ON_TX_FRAME);
			/* restore priority level and exit */
			splx(s);
			return;

		}
		
		/* dequeue packet from IP queue (only for CHDLC) */
		WAN_IFQ_DEQUEUE(tq, opkt);
		if (opkt){
			wan_skb_free(opkt);	/*m_freem(opkt);*/
		}

		/* update statistics and free the packet */
		card->outbytes += data_length;
 
	} /* while there're packets in the queue */

	/* restore interrupt level and exit */
	splx(s);
	return;
}



/****** Cisco HDLC Firmware Interface Functions *******************************/
/* 
 * ============================================================================
 * Read firmware code version.
 *	Put code version as ASCII string in str. 
 */
static int chdlc_read_version (sdla_t* card, char* str)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	char 	err;

	DEBUG_CMD("%s: Read CHDLC Version..\n", card->devname); 
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = READ_CHDLC_CODE_VERSION;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

#ifdef WANDEBUG_CHDLC
	chdlc_dump_mailbox(mbox);
#endif
	if (err != WAN_CMD_OK){ 
		chdlc_error(card, err, mbox);
	}else{
		if (str){ /* is not null */
			memcpy(str, mbox->wan_data, mbox->wan_data_len);
			str[mbox->wan_data_len] = '\0';
		}
	}
	return (err);
}

/* 
 * ============================================================================
 *  Configure CHDLC firmware.
 */
static int chdlc_configure (sdla_t* card, void* data)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		data_length = sizeof(CHDLC_CONFIGURATION_STRUCT);
	int 		err;

	DEBUG_INIT("%s: Setting CHDLC configuration...\n", card->devname);	
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = data_length;
	mbox->wan_command = SET_CHDLC_CONFIGURATION;
	memcpy(mbox->wan_data, data, data_length);
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
	        chdlc_error (card, err, mbox);
		DEBUG_EVENT("%s: Invalid value in offset %d\n", 
				card->devname, mbox->wan_data_len);  
	}
	return err;
}

/* 
 * ============================================================================
 * Set interrupt mode -- HDLC Version.
 */

static int chdlc_set_intr_mode (sdla_t* card, unsigned mode)
{
	wan_mbox_t*			mbox = &card->wan_mbox;
	CHDLC_INT_TRIGGERS_STRUCT*	int_data = 
			(CHDLC_INT_TRIGGERS_STRUCT*)mbox->wan_data;
	int err;

	DEBUG_CMD("%s: Setting interrupt mode(0x%x)...\n", card->devname, mode);
	WAN_MBOX_INIT(mbox);
	mbox->wan_command = SET_CHDLC_INTERRUPT_TRIGGERS;
	mbox->wan_data_len = sizeof(CHDLC_INT_TRIGGERS_STRUCT);
	int_data->CHDLC_interrupt_triggers	= mode;
	int_data->IRQ				= card->wandev.irq;
	int_data->interrupt_timer		= 1;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		chdlc_error (card, err, mbox);
	}
   
	return err;
}

/*
 * ============================================================================
 * Get interrupt mode -- HDLC Version
 */
#if 0
static int chdlc_read_intr_mode(sdla_t* card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err = 0;
     
	WAN_MBOX_INIT(mbox);
	mbox->wan_command = READ_CHDLC_INTERRUPT_TRIGGERS;
	mbox->wan_data_len = 0x0;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		chdlc_error (card, err, mbox);
	}else{
	    	CHDLC_INT_TRIGGERS_STRUCT int_data;
		memcpy(&int_data, mbox->wan_data, sizeof(CHDLC_INT_TRIGGERS_STRUCT));
	 	return int_data.CHDLC_interrupt_triggers;	
	}
     
	return err;
}
#endif

/* 
 * ============================================================================
 * Enable communications.
 */
static int chdlc_comm_enable (sdla_t* card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err = 0;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = ENABLE_CHDLC_COMMUNICATIONS;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		if (err == CHDLC_COMMS_ENABLED){
			DEBUG_CMD("%s: Communication already enabled\n", card->devname);
		}else{
			chdlc_error(card, err, mbox);
		}
	}else{
		DEBUG_CMD("%s: Communication enabled\n", card->devname);	
	}

	return err;
}

/* 
 * ============================================================================
 * Disable communications.
 */
static int chdlc_comm_disable (sdla_t* card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = DISABLE_CHDLC_COMMUNICATIONS;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		if (err == CHDLC_COMMS_DISABLED){
			DEBUG_CMD("%s: Communication already disabled\n", 
						card->devname); 
		}else{
			chdlc_error(card,err, mbox);
		}
	}else{
		DEBUG_CMD("%s: Communication disabled\n", card->devname);
	}
	return err;
}

/* 
 * ============================================================================
 * Read communication error statistics.
 */
static int chdlc_read_comm_err_stats (sdla_t* card, COMMS_ERROR_STATS_STRUCT* stats)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = READ_COMMS_ERROR_STATS;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		chdlc_error(card, err, mbox);
	}else{
		DEBUG_CMD("%s: Read communication error stat...ok\n", 
					card->devname);
		memcpy(stats, mbox->wan_data, sizeof(COMMS_ERROR_STATS_STRUCT));
	}
	return err;
}

/*
 ************************************************************************
 * Read CHDLC configureation.
 */
static int chdlc_read_configure(sdla_t *card, CHDLC_CONFIGURATION_STRUCT* cfg)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = READ_CHDLC_CONFIGURATION;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

	if(err != WAN_CMD_OK){
        	chdlc_error(card, err, mbox);
	}else{
		DEBUG_CMD("%s: Read CHDLC configuration...ok\n", card->devname);
		memcpy(cfg, mbox->wan_data, sizeof(CHDLC_CONFIGURATION_STRUCT));
	}
	return err;
}


/* 
 * ============================================================================
 * Read CHDLC operational statistics.
 */
static int chdlc_read_op_stats (sdla_t* card, CHDLC_OPERATIONAL_STATS_STRUCT* op_stats)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int	 	err;
	
	WAN_MBOX_INIT(mbox);
    	mbox->wan_data_len = 0;
    	mbox->wan_command = READ_CHDLC_OPERATIONAL_STATS;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
    	if (err != WAN_CMD_OK){
		chdlc_error(card, err, mbox);
    	}else{
		DEBUG_CMD("%s: Read CHDLC operational stat...ok\n", 
				card->devname);
		memcpy(op_stats, mbox->wan_data, sizeof(CHDLC_OPERATIONAL_STATS_STRUCT));
    	}
    	return err;
}

/* 
 * ============================================================================
 * Update communications error and general packet statistics.
 */
static int 
update_comms_stats(sdla_t* card)
{
	COMMS_ERROR_STATS_STRUCT	err_stats;
	CHDLC_OPERATIONAL_STATS_STRUCT 	op_stats;

	DEBUG_CMD("Updating communication statistics...\n");
	/* On the first timer interrupt, read the comms error statistics */
	if (card->update_comms_stats == 3){
		/* 1. On the first timer interrupt, update T1/E1 alarms 
		 * and PMON counters (only for T1/E1 card) (TE1) 
		 */
		/* TE1 Update T1/E1 alarms */
		if (IS_TE1_CARD(card)) {	
			card->wandev.fe_iface.read_alarm(&card->fe, 0); 
			/* TE1 Update T1/E1 perfomance counters */
			card->wandev.fe_iface.read_pmon(&card->fe);
		}else if (IS_56K_CARD(card)) {
			/* 56K Update CSU/DSU alarms */
			card->wandev.fe_iface.read_alarm(&card->fe, 1); 
		}
	}else if (card->update_comms_stats == 2){
		if (chdlc_read_comm_err_stats(card, &err_stats)){
			return 1;
		}
#if 0
		ifp->if_ierrors = 	err_stats.Rx_overrun_err_count + 
					err_stats.CRC_err_count +
					err_stats.Rx_abort_count +
					err_stats.Rx_dis_pri_bfrs_full_count +
					err_stats.sec_Tx_abort_count;
#endif
	}else{ /* on second timer interrupt, read the operational statistics */
		if (chdlc_read_op_stats(card, &op_stats)){
			return 1;
		}
#if 0
		ifp->if_ierrors +=	(op_stats.Rx_Data_discard_short_count +
					op_stats.Rx_Data_discard_long_count);
#endif
	}	
	return 0;
}


/*============================================================================
 * Send packet.
 *	Return:	0 - o.k.
 *		1 - no transmit buffers available
 */
static int chdlc_send (sdla_t* card, void* data, unsigned len)
{
	CHDLC_DATA_TX_STATUS_EL_STRUCT txbuf;

	card->hw_iface.peek(card->hw, card->u.c.txbuf_off, &txbuf, sizeof(txbuf));
	
	if (txbuf.opp_flag)
		return -EINVAL;
	
	card->hw_iface.poke(card->hw, txbuf.ptr_data_bfr, data, len);

	txbuf.frame_length = len;
	card->hw_iface.poke(card->hw, card->u.c.txbuf_off, &txbuf, sizeof(txbuf));
	
	txbuf.opp_flag = 1;		/* start transmission */
	card->hw_iface.poke_byte(card->hw, 
				 card->u.c.txbuf_off+offsetof(CHDLC_DATA_TX_STATUS_EL_STRUCT, opp_flag),
				 txbuf.opp_flag);

	/* Update transmit buffer control fields */
	card->u.c.txbuf_off += sizeof(txbuf);

	if (card->u.c.txbuf_off > card->u.c.txbuf_last_off)
		card->u.c.txbuf_off = card->u.c.txbuf_base_off;
	return 0;
}

/*============================================================================
 * Read TE1/56K Front end registers
 */
static unsigned char 
read_front_end_reg (void* card_id, unsigned short reg)
{
	sdla_t* 	card = (sdla_t*)card_id;
	wan_mbox_t*	mbox = &card->wan_mbox;
	char*		data = mbox->wan_data;
	int 		err = 0x0;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(FRONT_END_REG_STRUCT);
	mbox->wan_command = READ_FRONT_END_REGISTER;
	((FRONT_END_REG_STRUCT *)data)->register_number = (unsigned short)reg;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
        	chdlc_error(card,err, mbox);
    	}
	return (((FRONT_END_REG_STRUCT *)data)->register_value);
}

/*============================================================================
 * Write to TE1/56K Front end registers  
 */
static unsigned char 
write_front_end_reg (void* card_id, unsigned short reg, unsigned char value)
{
	sdla_t*		card = (sdla_t*)card_id;
	wan_mbox_t*	mbox = &card->wan_mbox;
	char*		data = mbox->wan_data;
	int 		err = 0x00;
	
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(FRONT_END_REG_STRUCT);
	mbox->wan_command = WRITE_FRONT_END_REGISTER;
	((FRONT_END_REG_STRUCT *)data)->register_number = (unsigned short)reg;
	((FRONT_END_REG_STRUCT *)data)->register_value = value;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
        	chdlc_error(card, err, mbox);
	}
	return err;
}

/*============================================================================
 * TE1 CHDLC enable hardware timer  
 */
static void hdlc_enable_timer(void* card_id)
{
	sdla_t* 			card = (sdla_t*)card_id;

	card->timer_int_enabled |= TMR_INT_ENABLED_TE;
	card->hw_iface.set_bit(card->hw, card->intr_perm_off, APP_INT_ON_TIMER);
    	return;
}


static int set_adapter_config (sdla_t* card)
{
	wan_mbox_t*			mbox = &card->wan_mbox;
	ADAPTER_CONFIGURATION_STRUCT*	cfg = (ADAPTER_CONFIGURATION_STRUCT*)mbox->wan_data; 
	int				err = 0x00;
	
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(ADAPTER_CONFIGURATION_STRUCT);
	mbox->wan_command = SET_ADAPTER_CONFIGURATION;
	card->hw_iface.getcfg(card->hw, SDLA_ADAPTERTYPE, &cfg->adapter_type);
	cfg->adapter_config = 0x00; 
	cfg->operating_frequency = 0x00; 
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
        	chdlc_error(card, err, mbox);
	}

	return (err);
}


/****** Firmware Error Handler **********************************************/

/* 
 * ============================================================================
 * Firmware error handler.
 *	This routine is called whenever firmware command returns non-zero
 *	return code.
 *
 * Return zero if previous command has to be cancelled.
 */
static int chdlc_error (sdla_t *card, int err, wan_mbox_t* mbox)
{
	switch (err){ 
	case WAN_CMD_TIMEOUT:
		DEBUG_EVENT("%s: command 0x%02x timed out!\n", 
					card->devname, mbox->wan_command);
		break;
	
	case S514_BOTH_PORTS_SAME_CLK_MODE:
		if (mbox->wan_command == SET_CHDLC_CONFIGURATION){
			DEBUG_EVENT("%s: Configure both ports for the same clock source\n",
						card->devname);
		}else{
			DEBUG_EVENT("%s: command 0x%02x returned 0x%02x!\n", 
						card->devname, mbox->wan_command, err);
		}
		break;
	
	default:
		DEBUG_EVENT("%s: command 0x%02x returned 0x%02x!\n", 
						card->devname, mbox->wan_command, err);
		break;
	}
	return 0;
}

/****** Interrupt Handlers **************************************************/

/* 
 * ============================================================================
 * Cisco HDLC interrupt service routine.
 */
static void wpc_isr (sdla_t* card)
{
	SHARED_MEMORY_INFO_STRUCT	flags;
	sdla_t*			my_card = NULL;
	int 			interrupt_serviced = 0;

	/* Check for which port the interrupt has been generated 
	 * Since Secondary Port is piggiback on the Primary the check be done here
	 */
	card->hw_iface.peek(card->hw, card->intr_type_off, &intr_type, sizeof(intr_type));
	if (!intr_type){
		/* Check for second port (piggiback) */
		if ((my_card = card->next) != NULL) {
    			card->hw_iface.peek(my_card->hw, my_card->intr_type_off, &intr_type, sizeof(intr_type));
			if (intr_type)
				card = my_card;
		}
	}
	card->in_isr = 1;

	/* On a 508 Card, if critical due to if_send Major Error */
	if (card->flags_off == 0){
		goto wpc_isr_done;
	}

       	card->hw_iface.peek(card->hw, card->intr_type_off, &intr_type, sizeof(intr_type));
	interrupt_serviced = 1;
	switch(intr_type){ 
	case RX_APP_INT_PEND:	/* 0x01: receive interrupt */
		rx_intr(card);
		break;

	case TX_APP_INT_PEND:	/* 0x02: transmit interrupt */
		card->hw_iface.clear_bit(card->hw, card->intr_perm_off, APP_INT_ON_TX_FRAME);
       
		chdlc_start(card);
		break;
             
	case COMMAND_COMPLETE_APP_INT_PEND:/* 0x04: cmd cplt */
		++ Intr_test_counter;
		break;

	case TIMER_APP_INT_PEND:
		timer_intr(card);
		break;

	case GLOBAL_EXCEP_COND_APP_INT_PEND:
		process_global_exception(card);
		/* Reset the 56k or T1/E1 front end exception condition */
		if(IS_56K_CARD(card) || IS_TE1_CARD(card)) {
			FRONT_END_STATUS_STRUCT FE_status;
			card->hw_iface.peek(card->hw, 
					    card->fe_status_off, 
					    &FE_status, 
					    sizeof(FRONT_END_STATUS_STRUCT));
			FE_status.opp_flag = 0x01;	
			card->hw_iface.poke(card->hw,
					    card->fe_status_off, 
					    &FE_status, 
					    sizeof(FRONT_END_STATUS_STRUCT));
		}
		break;

	case CHDLC_EXCEP_COND_APP_INT_PEND:	/* 0x20 */
		process_chdlc_exception(card);
		break;

	default:
		interrupt_serviced = 0;
		break;
	}

wpc_isr_done:
	if (!interrupt_serviced){ 
        	unsigned char str[5];
		card->hw_iface.peek(card->hw, card->flags_off, &flags, sizeof(SHARED_MEMORY_INFO_STRUCT));
        	DEBUG_EVENT("%s: spurious interrupt 0x%02x!\n", card->devname,
	        	        flags.interrupt_info_struct.interrupt_type);
		bcopy(flags.global_info_struct.codename, 
			str, 
			sizeof(flags.global_info_struct.codename));
		str[sizeof(flags.global_info_struct.codename)] = '\0';
		DEBUG_EVENT("%s: Code name:    %s\n", card->devname, str);
		bcopy(flags.global_info_struct.codeversion, 
			str, 
			sizeof(flags.global_info_struct.codeversion));
		str[sizeof(flags.global_info_struct.codeversion)] = '\0';
		DEBUG_EVENT("%s: Code version: %s\n", card->devname, str);
	}

	card->in_isr = 0;
	card->hw_iface.poke_byte(card->hw, card->intr_type_off, 0x00);

}

/* 
 * ============================================================================
 * Receive interrupt handler.
 */
static void rx_intr (sdla_t* card)
{
	SHARED_MEMORY_INFO_STRUCT    	flags;
	CHDLC_DATA_RX_STATUS_EL_STRUCT	rxbuf;
	struct mbuf*		 	ipkt = NULL; 
	unsigned long 		   	addr = 0;
	char* 			   	data_ptr = NULL;
	unsigned 		   	len = 0, api_hdr_len = 0;
	
	card->hw_iface.peek(card->hw, card->rxmb_off, &rxbuf, sizeof(CHDLC_DATA_RX_STATUS_EL_STRUCT));
	if (rxbuf.opp_flag != 0x01){ 
		unsigned char str[5];
	        card->hw_iface.peek(card->hw, card->flags_off, &flags, sizeof(SHARED_MEMORY_INFO_STRUCT));
		DEBUG_EVENT("%s: corrupted Rx buffer @ 0x%x, flag = 0x%02x!\n", 
				card->devname, (unsigned)card->rxmb_off, rxbuf.opp_flag);
		bcopy(flags.global_info_struct.codename, str,  
		      sizeof(flags.global_info_struct.codename));
		str[sizeof(flags.global_info_struct.codename)] = '\0';
	        DEBUG_EVENT("Code name:    %s\n", str);
		bcopy(flags.global_info_struct.codeversion, str,  
		      sizeof(flags.global_info_struct.codeversion));
		str[sizeof(flags.global_info_struct.codeversion)] = '\0';
	        DEBUG_EVENT("Code version: %s\n", str);

		/* 
		** 3.10.2000 RX_CORR
		** This case not suppose to be, but if it is, disable interrupt 
		** for current router. 
		*/
		DEBUG_EVENT("%s: Critical router failure...\n", card->devname);
		DEBUG_EVENT("Please contact your Sangoma represention!!\n");
		chdlc_set_intr_mode(card, 0);
		return;
	}
   
	addr = rxbuf.ptr_data_bfr;
	len  = rxbuf.frame_length;
	if (card->wandev.config_id == WANCONFIG_MLINK_PPP ||
	    card->wandev.config_id == WANCONFIG_MPPP){
		/* Take off two CRC bytes */
		len -= 2;
	}
	if (len <= 0){
		goto rx_intr_end;
	}

	api_hdr_len = 0;

	if (card->u.c.usedby == API){
		api_hdr_len = sizeof(api_rx_hdr_t);
	}

	/*
	** Check that we have anough space in rx buffer.
	*/
	if (len + api_hdr_len > MAX_PACKET_SIZE){
		DEBUG_EVENT("%s: Received packet is too long (%d bytes)!\n", 
				card->devname, len + api_hdr_len);
		goto rx_intr_end;
	}
    
	if (addr + len <= card->u.c.rx_top_off + 1){
		card->hw_iface.peek(card->hw, addr, &card->rx_data[api_hdr_len], len);
		addr += len;
	}else{
		int first_piece_len = card->u.c.rx_top_off + 1 - addr;
		int second_piece_len = len - first_piece_len;

		card->hw_iface.peek(card->hw, addr, &card->rx_data[api_hdr_len], first_piece_len);
		card->hw_iface.peek(card->hw, card->u.c.rx_base_off, 
                  	&card->rx_data[api_hdr_len]+first_piece_len, second_piece_len);
		addr = card->u.c.rx_base_off + second_piece_len; 
	}
	if (card->u.c.usedby == API){	/* Update API header */
		((api_rx_hdr_t*)card->rx_data)->error_flag = rxbuf.error_flag;
		((api_rx_hdr_t*)card->rx_data)->time_stamp = rxbuf.time_stamp;
	}
	len += api_hdr_len;
	data_ptr = card->rx_data;

	ipkt = sdla_get(NULL, data_ptr, len);

	if (ipkt == NULL){
		DEBUG_EVENT("%s: Failed to allocate mbuf (RX_INTR)!!\n", 
					card->devname);
		goto rx_intr_end;
	}

       	/* Ok, this is a regular packet - do what's said above */
	if (card->wandev.config_id == WANCONFIG_MLINK_PPP){
#ifdef NETGRAPH
		int error = 0;
		if (card->hook){
			NG_SEND_DATA_ONLY(error, 
					  card->hook, 
					  ipkt);
			card->inbytes += len;
		}
#else
		wan_skb_free(ipkt);
#endif
	} else if (card->wandev.config_id == WANCONFIG_CHDLC){
 		DEBUG_EVENT("%s: RX INTR Not supported!\n", card->devname); 
    	}

rx_intr_end:

	/* Release buffer element and calculate a pointer to the next one */
	rxbuf.opp_flag = 0x00; 
	card->hw_iface.poke(card->hw, card->rxmb_off, &rxbuf, sizeof(CHDLC_DATA_RX_STATUS_EL_STRUCT));
	card->rxmb_off += sizeof(CHDLC_DATA_RX_STATUS_EL_STRUCT);
	if (card->rxmb_off > card->u.c.rxbuf_last_off){
		card->rxmb_off = card->u.c.rxbuf_base_off;
	}
	return;
}

/* 
 * ============================================================================
 * Miscellaneous Functions
 *	- set_chdlc_config() used to set configuration options on the board
 */
static int set_chdlc_config(sdla_t *card)
{
	CHDLC_CONFIGURATION_STRUCT	cfg;

	bzero(&cfg, sizeof(CHDLC_CONFIGURATION_STRUCT));

	if (card->wandev.clocking){
		cfg.baud_rate = card->wandev.bps;
	}

	cfg.line_config_options = 
			(card->wandev.interface == WANOPT_RS232) ?
			INTERFACE_LEVEL_RS232 : INTERFACE_LEVEL_V35;

	if (card->wandev.config_id == WANCONFIG_MLINK_PPP){
		cfg.CHDLC_API_options	= DISCARD_RX_ERROR_FRAMES;
	}
	cfg.modem_config_options	= 0;     
	cfg.modem_status_timer		= 100;   
	cfg.CHDLC_protocol_options	= card->u.c.protocol_options;
	cfg.percent_data_buffer_for_Tx  = 50;
	cfg.CHDLC_statistics_options    = 0;
	cfg.max_CHDLC_data_field_length = card->wandev.mtu;
	cfg.transmit_keepalive_timer    = card->u.c.kpalv_tx;
	cfg.receive_keepalive_timer	= card->u.c.kpalv_rx;
	cfg.keepalive_error_tolerance   = card->u.c.kpalv_err;
	cfg.SLARP_request_timer		= card->u.c.slarp_timer;

	return chdlc_configure(card, &cfg);
}

/* 
 * ============================================================================
 * Process global exception condition
 */
static int process_global_exception(sdla_t *card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 		err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = READ_GLOBAL_EXCEPTION_CONDITION;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

	if (err != WAN_CMD_TIMEOUT){
		switch(err) {
		case EXCEP_MODEM_STATUS_CHANGE:
			if (IS_56K_CARD(card)){
				FRONT_END_STATUS_STRUCT FE_status;

				card->hw_iface.peek(card->hw,
						    card->fe_status_off, 
						    &FE_status, 
						    sizeof(FRONT_END_STATUS_STRUCT));
				card->fe.fe_param.k56_param.RR8_reg_56k = 
					FE_status.FE_U.stat_56k.RR8_56k;	
				card->fe.fe_param.k56_param.RRA_reg_56k = 
					FE_status.FE_U.stat_56k.RRA_56k;	
				card->fe.fe_param.k56_param.RRC_reg_56k = 
					FE_status.FE_U.stat_56k.RRC_56k;	
				card->wandev.fe_iface.read_alarm(&card->fe, 0);

				break;
			}

			if (IS_TE1_CARD(card)){
				/* TE1 T1/E1 interrupt */
				card->wandev.fe_iface.isr(&card->fe);
				break;
			}	
			DEBUG_EVENT("%s: Modem status change\n", card->devname);
			switch(mbox->wan_data[0] & (WAN_DCD_HIGH | WAN_CTS_HIGH)){ 
			case (WAN_DCD_HIGH):
				DEBUG_EVENT("%s: DCD high, CTS low\n", card->devname);
				break;
			case (WAN_CTS_HIGH):
				DEBUG_EVENT("%s: DCD low, CTS high\n", card->devname);       
				break;
			case ((WAN_DCD_HIGH | WAN_CTS_HIGH)):
				DEBUG_EVENT("%s: DCD high, CTS high\n", card->devname);
				break;
			default:
				DEBUG_EVENT("%s: DCD low, CTS low\n", card->devname);
				break;
			}
			break;

		case EXCEP_TRC_DISABLED:
			DEBUG_EVENT("%s: Line trace disabled\n", card->devname);
			break;

		case EXCEP_IRQ_TIMEOUT:
			DEBUG_EVENT("%s: IRQ timeout occurred\n", card->devname);
			break;

		case EXCEP_CSU_DSU_STATE_CHANGE:
			break;

		default:
			DEBUG_EVENT("%s: global exception (err=%02x)\n", card->devname, err);
			break;
		}
	}
	return 0;
}


/*
 * ============================================================================
 * Process chdlc exception condition
 */
static int process_chdlc_exception(sdla_t *card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 		err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = READ_CHDLC_EXCEPTION_CONDITION;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

	if (err != WAN_CMD_TIMEOUT){ 
		switch (err){ 
		case EXCEP_LINK_ACTIVE:
			DEBUG_EVENT("%s: CHDLC link became active ...\n", 
							card->devname);
			/* protocol is up - now if the slarp timer is zero, that 
			** means that IP address was set at the beginning and 
			** we are able for data transfers, so we will bring the  
			** interface up. If the slarp timer of station is not zero 
			** - we have to wait till we discover out IP address ... 
			*/
			if (card->u.c.slarp_timer == 0){
				DEBUG_EVENT("%s: Router has been brought up.\n", 
							card->devname); 
				wan_getcurrenttime(&card->u.c.router_up_time, NULL);
			} 	
			wanpipe_set_state(card, WAN_CONNECTED);
			break;

		case EXCEP_LINK_INACTIVE_MODEM:
			DEBUG_EVENT("%s: Link is down because of invalid modem state\n",
							card->devname);
			wanpipe_set_state(card, WAN_DISCONNECTED);
			unconfigure_ip(card);
			/* Start debugging */
			WAN_DEBUG_START(card);
			break;

		case EXCEP_LINK_INACTIVE_KPALV:
			DEBUG_EVENT("%s: link is down due to the station not receiving SLARP keepalive frames ...\n", 
							card->devname); 
			wanpipe_set_state(card, WAN_DISCONNECTED);
			DEBUG_EVENT("%s: Keepalive timer expired.\n", card->devname); 
			unconfigure_ip(card);
			/* Start debugging */
			WAN_DEBUG_START(card);
			break;

		case EXCEP_IP_ADDRESS_DISCOVERED:
			DEBUG_EVENT("%s: IP has been discovered ...\n", card->devname);
			if (configure_ip(card)){ 
				return -1;
			}
			break;

		case EXCEP_LOOPBACK_CONDITION:
			DEBUG_EVENT("%s: Loopback Condition Detected.\n", card->devname);
			break;

		case NO_CHDLC_EXCEP_COND_TO_REPORT:
			DEBUG_EVENT("%s: No exceptions reported.\n", card->devname);
			break;
		}
	}
	return 0;
}

/* 
 * ============================================================================
 * Configure IP from SLARP negotiation
 * This adds dynamic routes when SLARP has provided valid addresses
 */
static int configure_ip (sdla_t* card)
{
    return 0;
}

/* 
 * ============================================================================
 * Un-Configure IP negotiated by SLARP
 * This removes dynamic routes when the link becomes inactive.
 */
static int unconfigure_ip (sdla_t* card)
{
    return 0;
}


/* Finished
 * ============================================================================
 * Initialize Receive and Transmit Buffers.
 */

static void init_chdlc_tx_rx_buff(sdla_t* card)
{
	CHDLC_CONFIGURATION_STRUCT 		cfg;
	CHDLC_TX_STATUS_EL_CFG_STRUCT	tx_config;
	CHDLC_RX_STATUS_EL_CFG_STRUCT	rx_config;
	int err = 0;

	err = chdlc_read_configure(card, &cfg);
	if(err != WAN_CMD_OK){ 
		return;
	}

	card->hw_iface.peek(card->hw, cfg.ptr_CHDLC_Tx_stat_el_cfg_struct, 
		 &tx_config, 
		 sizeof(CHDLC_TX_STATUS_EL_CFG_STRUCT)); 
	card->hw_iface.peek(card->hw, cfg.ptr_CHDLC_Rx_stat_el_cfg_struct, 
		 &rx_config, 
		 sizeof(CHDLC_RX_STATUS_EL_CFG_STRUCT)); 

        /* Setup Head and Tails for buffers */
        card->u.c.txbuf_base_off = tx_config.base_addr_Tx_status_elements;
        card->u.c.txbuf_last_off = card->u.c.txbuf_base_off +  
		(tx_config.number_Tx_status_elements - 1) * sizeof(CHDLC_DATA_TX_STATUS_EL_STRUCT);

        card->u.c.rxbuf_base_off = rx_config.base_addr_Rx_status_elements;
        card->u.c.rxbuf_last_off = card->u.c.rxbuf_base_off +
		(rx_config.number_Rx_status_elements - 1) * sizeof(CHDLC_DATA_RX_STATUS_EL_STRUCT);

        /* Set up next pointer to be used */
        card->u.c.txbuf_off = tx_config.next_Tx_status_element_to_use;
        card->rxmb_off = rx_config.next_Rx_status_element_to_use;

	/* Setup Actual Buffer Start and end addresses */
	card->u.c.rx_base_off = rx_config.base_addr_Rx_buffer;
	card->u.c.rx_top_off  = rx_config.end_addr_Rx_buffer;
	return;
}

#ifdef CHDLC_OLD_INTR_TEST
/* 
 * ============================================================================
 * Perform Interrupt Test by running READ_CHDLC_CODE_VERSION command MAX_INTR
 * _TEST_COUNTER times.
 */
static int chdlc_intr_test(sdla_t* card)
{
	int 	err = 0, i;

	Intr_test_counter = 0;

	err = chdlc_set_intr_mode(card, APP_INT_ON_COMMAND_COMPLETE); 
	if (err == WAN_CMD_OK){ 
		wan_mbox_t*	mbox = &card->wan_mbox;
		WAN_MBOX_INIT(mbox);
		for (i = 0; i < MAX_INTR_TEST_COUNTER; i ++){ 
			mbox->wan_data_len  = 0;
			mbox->wan_command = READ_CHDLC_CODE_VERSION;
			err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
			if (err != WAN_CMD_OK){ 
				chdlc_error(card, err, mbox);
			}
		}
		err = chdlc_set_intr_mode(card, 0x0);
	}

	if (err || (Intr_test_counter < MAX_INTR_TEST_COUNTER)){
		DEBUG_EVENT("%s: Interrupt test failed (%d)\n",
					card->devname, Intr_test_counter);
		DEBUG_EVENT("%s: Please choose another interrupt\n",
					card->devname);
		return -EIO;
	}
	DEBUG_EVENT("%s: Interrupt test passed(%d)\n",
					card->devname, Intr_test_counter);
	return 0;
}
#endif


/* Finished
 * ============================================================================
 * Timer interrupt handler.
 */
void timer_intr(sdla_t *card)
{
	/* process a udp call if pending */
#ifndef CHDLC_OLD_INTR_TEST
	if (card->timer_int_enabled & TMR_INT_ENABLED_TEST){
		Intr_test_counter++;
    		if (Intr_test_counter >= MAX_NEW_INTR_TEST_COUNTER){
    			DEBUG_EVENT("%s: Interrupt test passed(%d)\n",
					card->devname, Intr_test_counter);
        		card->timer_int_enabled &= ~TMR_INT_ENABLED_TEST;
			card->timer_int_enabled |= TMR_INT_ENABLED_CONFIG;
		}
	}
#endif
 
	/* TE1 T1/E1 timer interrupt */
	if (card->timer_int_enabled & TMR_INT_ENABLED_TE){
		card->wandev.fe_iface.polling(&card->fe);
        	card->timer_int_enabled &= ~TMR_INT_ENABLED_TE;
	}

	/* Chdlc configure */
	if (card->timer_int_enabled & TMR_INT_ENABLED_CONFIG){
		chdlc_config(card);
        	card->timer_int_enabled &= ~TMR_INT_ENABLED_CONFIG;
	}

	/* read communicationstatistics if required */
	if (card->timer_int_enabled & TMR_INT_ENABLED_UPDATE){
		update_comms_stats(card);
		if (!(--card->update_comms_stats)){
        		card->timer_int_enabled &= ~TMR_INT_ENABLED_UPDATE;
		}
	}
	/* Only disable the timer interrupt if there are no udp or statistics 
	 * updates pending
	 */
	if (!card->timer_int_enabled){
		card->hw_iface.clear_bit(card->hw, card->intr_perm_off, APP_INT_ON_TIMER);
	}
}



/*
****************************************************************************
**
**
**
*/
static int chdlc_debugging(sdla_t* card)
{
	CHDLC_OPERATIONAL_STATS_STRUCT	op_stats;
	static unsigned long	rx_keepalives = 0;

	chdlc_read_op_stats(card, &op_stats);
	if (card->wan_debugging_state == WAN_DEBUGGING_CONT){
		rx_keepalives	= op_stats.CHDLC_SLARP_REPLY_Rx_count;
		card->wan_debugging_state = WAN_DEBUGGING_PROTOCOL;
		/*
		** Check number of keepalives frames after 15 seconds.
		*/
		return 15;
	}

	if (op_stats.CHDLC_SLARP_REPLY_Rx_count != rx_keepalives){
		if (card->wan_debug_last_msg != WAN_DEBUG_CHDLC_UNKNWN_MSG){
			DEBUG_EVENT("%s: Unknown CHDLC problem!\n",
						card->devname);
			DEBUG_EVENT("%s:    Contact your dealer.\n",
						card->devname);
		}
		card->wan_debug_last_msg = WAN_DEBUG_CHDLC_UNKNWN_MSG;
	}else{
		if (card->wan_debug_last_msg != WAN_DEBUG_CHDLC_KPLV_MSG){
			DEBUG_EVENT("%s: No replies to Keepalive packets!\n",
						card->devname);
			DEBUG_EVENT("%s:    Check remote router or ISP.\n",
						card->devname);
		}
		card->wan_debug_last_msg = WAN_DEBUG_CHDLC_KPLV_MSG;
	}
	return 0;
}

static unsigned long chdlc_crc_frames(sdla_t* card)
{
	COMMS_ERROR_STATS_STRUCT	stats;

 	if (!chdlc_read_comm_err_stats(card, &stats)){
		return stats.CRC_err_count;	
	}
	return 0;
}

static unsigned long chdlc_abort_frames(sdla_t * card)
{
	COMMS_ERROR_STATS_STRUCT	stats;

 	if (!chdlc_read_comm_err_stats(card, &stats)){
		return stats.Rx_abort_count;	
	}
	return 0;
}

static unsigned long chdlc_tx_underun_frames(sdla_t* card)
{
	COMMS_ERROR_STATS_STRUCT	stats;

 	if (!chdlc_read_comm_err_stats(card, &stats)){
		return stats.sec_Tx_abort_msd_Tx_int_count;	
	}

	return 0;
}


#ifdef WANDEBUG_CHDLC
static void 
chdlc_dump_mailbox(wan_mbox_t *mbox)
{
    DEBUG_EVENT("Opp Flag      = 0x%x\n", mbox->wan_opp_flag);
    DEBUG_EVENT("Command       = 0x%x\n", mbox->wan_command);
    DEBUG_EVENT("Buffer Length = 0x%x\n", mbox->wan_data_len);
    DEBUG_EVENT("Return code   = 0x%x\n", mbox->wan_return_code);
}
#endif

static void chdlc_down(sdla_t* card)
{
	chdlc_set_intr_mode(card, 0x00);
	chdlc_comm_disable(card);
}

static int chdlc_up(sdla_t* card)
{
	/* Set interrupt mode and mask */
	if (chdlc_set_intr_mode(card, 
                        APP_INT_ON_RX_FRAME | APP_INT_ON_GLOBAL_EXCEP_COND |
                        APP_INT_ON_TX_FRAME | APP_INT_ON_CHDLC_EXCEP_COND | 
	                APP_INT_ON_TIMER)) 
	{
		return -EIO;
	}

	/* Mask the Transmit and Timer interrupt */
	card->hw_iface.clear_bit(
				card->hw, 
				card->intr_perm_off, 
				(APP_INT_ON_TX_FRAME | APP_INT_ON_TIMER));

	/* Enable communications */
	if (chdlc_comm_enable(card)){ 
		return -EIO;
	}
	return 0;
}

static void hdlc_handle_front_end_state(void* card_id)
{
	/* FIXME: Add handling front end status changes */

	return;
}

/****** End ****************************************************************/
